from typing import Optional
from .errors import PoolMaxConnectionsError
from .conn import PySQLXEngineSync


class PySQLXEnginePoolSync:
	__slots__ = ("uri", "max_connections", "_pool_len")

	def __init__(self, uri: str, max_connections: Optional[int] = None):
		PySQLXEngineSync(uri)
		self.uri = uri
		assert isinstance(max_connections, (int, None)), "max_connections must be an integer or None"
		self.max_connections: int = max_connections
		self._pool_len: int = 0

	def __enter__(self) -> PySQLXEngineSync:
		return self.new_connection()

	def __exit__(self, exc_type, exc, exc_tb): ...

	def new_connection(self) -> PySQLXEngineSync:
		if self.max_connections is not None and self._pool_len >= self.max_connections:
			raise PoolMaxConnectionsError()

		self._pool_len += 1
		conn = PySQLXEngineSync(self.uri)
		conn.connect()
		return conn



from functools import lru_cache
import json

from .const import LOG_CONFIG

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import JsonLexer, PythonLexer
from pygments.lexers.sql import SqlLexer


def _colorizer_json_message(dumps: str):
	return highlight(dumps, JsonLexer(), TerminalFormatter())


def fe_json(data: dict) -> str:
	"""create a error message format for python code"""
	dumps = json.dumps(data, indent=2)
	if LOG_CONFIG.PYSQLX_USE_COLOR:
		return "\n" + _colorizer_json_message(dumps)
	return "\n" + dumps


def _colorizer_py_code_message(message: str):
	return highlight(message, PythonLexer(), TerminalFormatter())


def fe_py(message: str) -> str:
	"""create a error message format for python code"""
	if LOG_CONFIG.PYSQLX_USE_COLOR:
		return _colorizer_py_code_message(message)
	return message


@lru_cache(maxsize=None)
def _colorizer_sql_message(sql: str):
	return highlight(sql, SqlLexer(), TerminalFormatter())


def fe_sql(sql: str) -> str:
	"""create a error message format for python code"""
	if LOG_CONFIG.PYSQLX_USE_COLOR:
		return _colorizer_sql_message(sql)
	return sql


def isolation_error_message():
	return f"""
    the isolation_level must be a valid value.

    possible values:
        >> ReadUncommitted
        >> ReadCommitted
        >> RepeatableRead
        >> Snapshot
        >> Serializable

    example of use:
        >> {fe_py('isolation_level = "ReadUncommitted"')}
    """


def not_connected_error_message():
	return f"""
        not connected to the database.

        before using the methods, you must connect to the database.
            
        example of use:
            Async:
                {fe_py('''uri = "postgresql://user:pass@host:port/db?schema=sample"
                db = PySQLXEngine(uri=uri)
                await db.connect()
                ''')}
            Sync:
                {fe_py('''uri = "postgresql://user:pass@host:port/db?schema=sample"
                db = PySQLXEngineSync(uri=uri)
                db.connect()
                ''')}
    """


def model_parameter_error_message():
	return f"""
        model parameter must be a subclass of BaseRow

        try importing the BaseRow class from the pysqlx_engine package.

        example of use:
            {fe_py('''
            
            # import the BaseRow class
            from pysqlx_engine import BaseRow

            class MyModel(BaseRow): # <- subclass of BaseRow
                id: int
                name: str
    
            # async
            db = PySQLXEngine(uri="postgresql://user:pass@host:port/db?schema=sample")
            await db.connect()
            await db.query(sql="SELECT 1 AS id, 'Rian' AS name", model=MyModel) # <- using how model parameter

            # sync
            db = PySQLXEngineSync(uri="postgresql://user:pass@host:port/db?schema=sample")
            db.connect()
            db.query(sql="SELECT 1 AS id, 'Rian' AS name", model=MyModel) # <- using how model parameter
            ''')
            }
        """


def sql_type_error_message():
	return f"""
        the queries and statements must be a valid string.

        you can not use a type other than string.

        example of use:
            {fe_py('''
            
            # async
            db = PySQLXEngine(uri="postgresql://user:pass@host:port/db?schema=sample")
            await db.connect()
            await db.query(sql="SELECT 1 AS id, 'Rian' AS name") # <- Literal string
            # or
            sql = "SELECT 1 AS id, 'Rian' AS name"
            await db.query(sql=sql) # <- Variable string


            # sync
            db = PySQLXEngineSync(uri="postgresql://user:pass@host:port/db?schema=sample")
            db.connect()
            db.query(sql="SELECT 1 AS id, 'Rian' AS name") # <- Literal string
            # or
            sql = "SELECT 1 AS id, 'Rian' AS name"
            db.query(sql=sql) # <- Variable string

            ''')}
    """


def parameters_type_error_message():
	return f"""
        the parameters must be a valid dict.

        you can not use a type other than dict[str, any].

        * dict key must be a valid string.
        * dict value can be a types: bool, str, int, list, dict, tuple, UUID, time, date, datetime, float, bytes, Decimal

        Python types vs SQL types:
            bool     -> bool/bit/boolean/tinyint/etc
            str      -> varchar/text/nvarchar/char/etc
            int      -> int/integer/smallint/bigint/tinyint/etc
            list     -> json/jsonb/nvarchar/varchar/string/etc
            dict     -> json/jsonb/nvarchar/varchar/string/etc
            tuple    -> array(Postgres Native)
            UUID     -> uuid/varchar/text/nvarchar/etc
            time     -> time/nvarchar/varchar/string/etc
            date     -> date/nvarchar/varchar/string/etc
            datetime -> timestamp/timestamptz/datetime/datetime2/nvarchar/varchar/string/etc
            float    -> float/real/numeric
            bytes    -> bytea/binary/varbinary
            Decimal  -> decimal/numeric
            None     -> null


        example of use:
            {fe_py('''
            
            # ===== async
            db = PySQLXEngine(uri="postgresql://user:pass@host:port/db?schema=sample")
            await db.connect()

            # Literal string with parameters
            await db.query(sql="SELECT :id AS id, :name AS name", parameters={"id": 1, "name": "Rian"})
            
            # or

            # Variable string with parameters
            sql = "SELECT :id AS id, :name AS name"
            parameters = {"id": 1, "name": "Rian"}
            await db.query(sql=sql, parameters=parameters) 


            # ===== sync
            db = PySQLXEngineSync(uri="postgresql://user:pass@host:port/db?schema=sample")
            db.connect()

            # Literal string with parameters
            db.query(sql="SELECT :id AS id, :name AS name", parameters={"id": 1, "name": "Rian"})
            
            # or
            
            # Variable string with parameters
            sql = "SELECT :id AS id, :name AS name"
            parameters = {"id": 1, "name": "Rian"}
            db.query(sql=sql, parameters=parameters) 

            ''')}
    """



from typing import Optional

from .._core.parser import MyModel  # import necessary using _core to not subscribe default parser
from .const import ISOLATION_LEVEL
from .util import force_sync
from .aconn import PySQLXEngine as _PySQLXEngine


class PySQLXEngineSync(_PySQLXEngine):
	uri: str
	connected: bool

	def __init__(self, uri: str):
		super().__init__(uri=uri)

	def __enter__(self):
		self.connect()
		return self

	def __exit__(self, exc_type, exc, exc_tb):
		self.close()

	@force_sync
	async def connect(self):
		await (super()).connect()

	@force_sync
	async def close(self):
		await (super()).close()

	@force_sync
	async def raw_cmd(self, sql: str):
		return await (super()).raw_cmd(sql=sql)

	@force_sync
	async def query(self, sql: str, parameters: Optional[dict] = None, model: Optional[MyModel] = None):
		return await (super()).query(sql=sql, parameters=parameters, model=model)

	@force_sync
	async def query_as_dict(self, sql: str, parameters: Optional[dict] = None):
		return await (super()).query_as_dict(sql=sql, parameters=parameters)

	@force_sync
	async def query_first(self, sql: str, parameters: Optional[dict] = None, model: Optional[MyModel] = None):
		return await (super()).query_first(sql=sql, parameters=parameters, model=model)

	@force_sync
	async def query_first_as_dict(self, sql: str, parameters: Optional[dict] = None):
		return await (super()).query_first_as_dict(sql=sql, parameters=parameters)

	@force_sync
	async def execute(self, sql: str, parameters: Optional[dict] = None):
		return await (super()).execute(sql=sql, parameters=parameters)

	@force_sync
	async def set_isolation_level(self, isolation_level: ISOLATION_LEVEL):
		await (super()).set_isolation_level(isolation_level=isolation_level)

	@force_sync
	async def begin(self):
		await (super()).begin()

	@force_sync
	async def commit(self):
		await (super()).commit()

	@force_sync
	async def rollback(self):
		await (super()).rollback()

	@force_sync
	async def start_transaction(self, isolation_level: Optional[ISOLATION_LEVEL] = None):
		await (super()).start_transaction(isolation_level=isolation_level)



from abc import ABC, abstractmethod

from typing_extensions import TypeVar

from .const import PROVIDER

T = TypeVar("T")


class AbstractDatabaseType(ABC):
	"""
	Abstract database type

	You can create a custom type by inheriting this class.

	Implement the `convert` method to convert the value to the database type.

	"""

	@abstractmethod
	def __init__(self, value: T): ...

	@abstractmethod
	def convert(self, provider: PROVIDER, field: str = "") -> T: ...



import json
from datetime import date, datetime, time
from decimal import Decimal
from uuid import UUID

from .errors import ParameterInvalidJsonValueError


class PySQLXJsonEnconder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, bytes):
			return obj.hex()

		elif isinstance(obj, (UUID, time, date, datetime, Decimal)):
			return str(obj)
		try:
			return super().default(obj)
		except TypeError as err:
			raise ParameterInvalidJsonValueError(
				typ_from=type(obj),
				typ_to="json",
				details=str(err),
			)



from types import TracebackType
from typing import Any, Dict, List, Optional, Type, Union, overload

from .._core.parser import BaseRow, MyModel, DictParam  # import necessary using _core to not subscribe default parser
from .const import ISOLATION_LEVEL

class PySQLXEngineSync:
	"""
	## Description

	PySQLXEngine is an engine to run pure sql, but you have flexibility to use how you want.

	All SQL that is executed using the PySQLXEngine is atomic; that is, only one statement is performed at a time.

	Only the first one will be completed if you send an Insert and a select.
	This is one of the ways to deal with SQL ingestion.

	By default the `BEGIN`, `COMMIT` and `ROLLBACK` is automatic (CASE DATABASE NOT NEED REQUIRES ISOLATION FIRST), if the sql is valid, is committed, if not, is rolled back.

	But you can use the `BEGIN` and `COMMIT` or `ROLLBACK` to control the transaction.

	---

	### Arguments:

	    `uri(str)`: uri of the database, example `postgresql://user:pass@host:port/db?schema=sample`

	---

	### Examples:


	##### PostgreSQL
	```python
	from pysqlx_engine import PySQLXEngineSync

	uri = "postgresql://user:pass@host:port/db?schema=sample"
	db = PySQLXEngineSync(uri=uri)
	db.connect()
	```
	---
	##### MySQL
	```python
	from pysqlx_engine import PySQLXEngineSync

	uri = "mysql://user:pass@host:port/db?schema=sample"
	db = PySQLXEngineSync(uri=uri)
	db.connect()
	```
	---
	##### Microsoft SQL Server
	```python
	from pysqlx_engine import PySQLXEngineSync

	uri = "sqlserver://host:port;initial catalog=sample;user=sa;password=pass;"
	db = PySQLXEngineSync(uri=uri)
	db.connect()
	```
	---
	##### SQLite
	```python
	from pysqlx_engine import PySQLXEngineSync

	uri = "sqlite:./dev.db"
	db = PySQLXEngineSync(uri=uri)
	db.connect()
	```
	"""

	__slots__ = ["uri", "connected", "_conn", "_provider"]

	uri: str
	connected: bool

	def __init__(self, uri: str) -> "None": ...
	def __del__(self):
		"""
		## Description

		Automatically close the connection when the object is deleted.
		"""
		...
	def is_healthy(self) -> "bool":
		"""
		## Description

		Check if the connection is healthy.

		Returns false, if connection is considered to not be in a working state.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``bool``

		    * Raises: ``None``

		"""
		...
	def requires_isolation_first(self) -> "bool":
		"""
		## Description

		Returns `True` if the connection requires isolation first, `False` otherwise.

		This is used to determine if the connection should be isolated before executing a query.

		For example, sqlserver requires isolation before executing a statement using begin in some cases.

		Signals if the isolation level SET needs to happen before or after the BEGIN.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``bool``

		    * Raises: ``None``

		---

		### Extra documentation:
		    * [MSSQL](https://learn.microsoft.com/en-us/sql/t-sql/language-elements/transaction-isolation-levels)
		    * [Postgres](https://www.postgresql.org/docs/current/sql-set-transaction.html)
		    * [MySQL](https://dev.mysql.com/doc/refman/8.0/en/innodb-transaction-isolation-levels.html)
		    * [SQLite](https://www.sqlite.org/isolation.html)
		"""
		...
	def __enter__(self) -> "PySQLXEngineSync":
		"""
		## Description

		Open a connection to the database. using `with`.
		"""
		...
	def __exit__(
		self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], exc_tb: Optional[TracebackType]
	): ...
	def connect(self) -> "None":
		"""
		## Description

		Each connection instance is lazy; only after ``.connect()`` is the database checked and the connection established.

		When you use ``async with`` the connection is automatically opened and closed.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``None``

		    * Raises: ``ConnectError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()
		```
		"""
		...
	def close(self) -> "None":
		"""
		## Description

		It's a good idea to close the connection, but PySQLXEngine has the core in Rust;
		closing is automatic when your code leaves the context.

		Even if you don't close the connection, don't worry; when the process ends automatically,
		the connections will be closed, so the database doesn't have an idle connection.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``None``

		    * Raises: ``None``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()
		    db.close()
		```
		"""
		...
	def raw_cmd(self, sql: str) -> "None":
		"""
		## Description
		Run a command in the database, for queries that can't be run using prepared statements(queries/execute).

		---

		### Helper

		    * Arguments:

		        ``sql(str)``: sql to be executed.

		    * Returns: ``None``

		    * Raises: ``RawCmdError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    db.raw_cmd(sql="SET TRANSACTION ISOLATION LEVEL READ COMMITTED;")

		```
		"""
		...
	# all
	@overload
	def query(self, sql: str, parameters: Optional[DictParam] = None) -> Union[List[BaseRow], List]:
		"""
		## Description

		Returns all rows from query result as``BaseRow list``, ``MyModel list`` or ``empty list``.

		---

		### Helper
		    * Arguments:

		        ``sql(str)``: sql query to be executed.

		        ``parameters(dict)``: (default is None) parameters must be a dictionary with the name of the parameter and the value.

		        ``model(BaseRow)``: (default is None) is your model that inherits from BaseRow.

		    * Returns:

		        ``List[BaseRow] | List[MyModel] | List``: BaseRow list, MyModel list or empty list.

		    * Raises: ``QueryError``|``TypeError`` | ``ParameterInvalidProviderError``|``ParameterInvalidValueError``|``ParameterInvalidJsonValueError``

		---

		### Parameters Helper

		Parameters are built into SQL at the application level; that is, the SQL and separate parameters are not sent to the database;
		although most databases support this type of operation, the PySQLXEngine does it before calling the database to avoid possible incompatibilities.
		This allows you to show the precompiled queries and send only raw SQL while maintaining minimal consistency across types.

		#### SQL with parameters syntax
		    * SQL: ``SELECT * FROM table WHERE id = :id``
		    * Parameters: ``{"id": 1}``

		#### Parameters(dict):

		    * dict ``key`` must be a valid string.
		    * dict ``value`` can be one of the types: (
		        ``bool``,
		        ``bytes``,
		        ``date``,
		        ``datetime``,
		        ``Decimal``,
		        ``dict``,
		        ``Enum``, # Enum must be a subclass of enum.Enum
		        ``float``,
		        ``int``,
		        ``list``,
		        ``str``,
		        ``time``,
		        ``tuple``,
		        ``UUID``,
		        ``None``
		    )

		#### Python types vs SQL types:

		    [Documentation](https://carlos-rian.github.io/pysqlx-engine/type_mappings/)

		```
		    * bool     -> bool|bit|boolean|tinyint|etc
		    * bytes    -> bytea|binary|varbinary|etc
		    * date     -> date|nvarchar|varchar|string|etc
		    * datetime -> timestamp|timestamptz|datetime|datetime2|nvarchar|varchar|string|etc
		    * Decimal  -> decimal|numeric|etc
		    * dict     -> json|jsonb|nvarchar|varchar|string|etc
		    * float    -> float|real|numeric|etc
		    * int      -> int|integer|smallint|bigint|tinyint|etc
		    * list     -> json|jsonb|nvarchar|varchar|string|etc
		    * str      -> varchar|text|nvarchar|char|etc
		    * time     -> time|nvarchar|varchar|string|etc
		    * tuple    -> array(Postgres Native), another database: error.
		    * UUID     -> uuid|varchar|text|nvarchar|etc
		    * Enum     -> varchar|text|nvarchar|etc
		    * None     -> null
		```

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    result = db.query("SELECT 1 as id, 'rian' as name")
		    print(result)
		    # output -> [BaseRow(id=1, name='rian')]

		    result = db.query(sql="SELECT 1 as :id, 'rian' as name", parameters={"id": 1})
		    print(result)
		    # output -> [BaseRow(id=1, name='rian')]

		    db.close()

		```
		"""
		...
	@overload
	def query(
		self, sql: str, parameters: Optional[DictParam] = None, model: Optional[Type["MyModel"]] = None
	) -> Union[List[Type["MyModel"]], List]: ...
	# dict
	def query_as_dict(self, sql: str, parameters: Optional[DictParam] = None) -> Union[List[Dict[str, Any]], List]:
		"""
		## Description

		Returns all rows from query result as ``dict list`` or ``empty list``.

		---

		### Helper
		    * Arguments:

		        ``sql(str)``: sql query to be executed

		        ``parameters(dict)``: (default is None) parameters must be a dictionary with the name of the parameter and the value.

		    * Returns:

		        ``List[Dict[str, Any]] | List``: dict list or empty list.

		    * Raises: ``QueryError``|``TypeError`` | ``ParameterInvalidProviderError``|``ParameterInvalidValueError``|``ParameterInvalidJsonValueError``

		---

		### Parameters Helper

		Parameters are built into SQL at the application level; that is, the SQL and separate parameters are not sent to the database;
		although most databases support this type of operation, the PySQLXEngine does it before calling the database to avoid possible incompatibilities.
		This allows you to show the precompiled queries and send only raw SQL while maintaining minimal consistency across types.

		#### SQL with parameters syntax
		    * SQL: ``SELECT * FROM table WHERE id = :id``
		    * Parameters: ``{"id": 1}``

		#### Parameters(dict):

		    * dict ``key`` must be a valid string.
		    * dict ``value`` can be one of the types: (
		        ``bool``,
		        ``bytes``,
		        ``date``,
		        ``datetime``,
		        ``Decimal``,
		        ``dict``,
		        ``Enum``, # Enum must be a subclass of enum.Enum
		        ``float``,
		        ``int``,
		        ``list``,
		        ``str``,
		        ``time``,
		        ``tuple``,
		        ``UUID``,
		        ``None``
		    )

		#### Python types vs SQL types:

		    [Documentation](https://carlos-rian.github.io/pysqlx-engine/type_mappings/)

		```
		    * bool     -> bool|bit|boolean|tinyint|etc
		    * bytes    -> bytea|binary|varbinary|etc
		    * date     -> date|nvarchar|varchar|string|etc
		    * datetime -> timestamp|timestamptz|datetime|datetime2|nvarchar|varchar|string|etc
		    * Decimal  -> decimal|numeric|etc
		    * dict     -> json|jsonb|nvarchar|varchar|string|etc
		    * float    -> float|real|numeric|etc
		    * int      -> int|integer|smallint|bigint|tinyint|etc
		    * list     -> json|jsonb|nvarchar|varchar|string|etc
		    * str      -> varchar|text|nvarchar|char|etc
		    * time     -> time|nvarchar|varchar|string|etc
		    * tuple    -> array(Postgres Native), another database: error.
		    * UUID     -> uuid|varchar|text|nvarchar|etc
		    * Enum     -> varchar|text|nvarchar|etc
		    * None     -> null
		```

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    result = db.query_as_dict(sql="SELECT 1 as id, 'rian' as name")
		    print(result)
		    # output -> [{'id': 1, 'name': 'rian'}]

		    result = db.query_as_dict(sql="SELECT 1 as :id, 'rian' as name", parameters={"id": 1})
		    print(result)
		    # output -> [{'id': 1, 'name': 'rian'}]

		    db.close()
		```
		"""
		...
	# fisrt
	@overload
	def query_first(self, sql: str, parameters: DictParam = None) -> Union[BaseRow, None]:
		"""
		## Description

		Returns first row from query result as ``BaseRow``, ``MyModel`` or ``None``.

		---

		### Helper
		    * Arguments:

		        ``sql(str)``: sql query to be executed.

		        ``parameters(dict)``: (default is None) parameters must be a dictionary with the name of the parameter and the value.

		        ``model(BaseRow)``: (default is None) is your model that inherits from BaseRow.

		    * Returns:

		        ``BaseRow | MyModel | None``: BaseRow, MyModel or None if no rows are found.

		    * Raises: ``QueryError``|``TypeError`` | ``ParameterInvalidProviderError``|``ParameterInvalidValueError``|``ParameterInvalidJsonValueError``

		---

		### Parameters Helper

		Parameters are built into SQL at the application level; that is, the SQL and separate parameters are not sent to the database;
		although most databases support this type of operation, the PySQLXEngine does it before calling the database to avoid possible incompatibilities.
		This allows you to show the precompiled queries and send only raw SQL while maintaining minimal consistency across types.

		#### SQL with parameters syntax
		    * SQL: ``SELECT * FROM table WHERE id = :id``
		    * Parameters: ``{"id": 1}``

		#### Parameters(dict):

		    * dict ``key`` must be a valid string.
		    * dict ``value`` can be one of the types: (
		        ``bool``,
		        ``bytes``,
		        ``date``,
		        ``datetime``,
		        ``Decimal``,
		        ``dict``,
		        ``Enum``, # Enum must be a subclass of enum.Enum
		        ``float``,
		        ``int``,
		        ``list``,
		        ``str``,
		        ``time``,
		        ``tuple``,
		        ``UUID``,
		        ``None``
		    )

		#### Python types vs SQL types:

		    [Documentation](https://carlos-rian.github.io/pysqlx-engine/type_mappings/)

		```
		    * bool     -> bool|bit|boolean|tinyint|etc
		    * bytes    -> bytea|binary|varbinary|etc
		    * date     -> date|nvarchar|varchar|string|etc
		    * datetime -> timestamp|timestamptz|datetime|datetime2|nvarchar|varchar|string|etc
		    * Decimal  -> decimal|numeric|etc
		    * dict     -> json|jsonb|nvarchar|varchar|string|etc
		    * float    -> float|real|numeric|etc
		    * int      -> int|integer|smallint|bigint|tinyint|etc
		    * list     -> json|jsonb|nvarchar|varchar|string|etc
		    * str      -> varchar|text|nvarchar|char|etc
		    * time     -> time|nvarchar|varchar|string|etc
		    * tuple    -> array(Postgres Native), another database: error.
		    * UUID     -> uuid|varchar|text|nvarchar|etc
		    * Enum     -> varchar|text|nvarchar|etc
		    * None     -> null
		```

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    result = db.query_first("SELECT 1 as id, 'rian' as name")
		    print(result)
		    # output -> BaseRow(id=1, name='rian')

		    result = db.query_first(sql="SELECT 1 as :id, 'rian' as name", parameters={"id": 1})
		    print(result)
		    # output -> BaseRow(id=1, name='rian')

		    db.close()

		```
		"""
		...
	@overload
	def query_first(
		self, sql: str, parameters: DictParam = None, model: Type["MyModel"] = None
	) -> Union[Type["MyModel"], None]: ...
	# dict
	def query_first_as_dict(self, sql: str, parameters: Optional[DictParam] = None) -> Optional[Dict[str, Any]]:
		"""
		## Description

		Returns first row from query result as ``dict`` or ``None``.

		---

		### Helper
		    * Arguments:

		        ``sql(str)``: sql query to be executed.

		        ``parameters(dict)``: (default is None) parameters must be a dictionary with the name of the parameter and the value.

		    * Returns:

		        ``Dict[str, Any] | None``: dict or None.

		    * Raises: ``QueryError``|``TypeError`` | ``ParameterInvalidProviderError``|``ParameterInvalidValueError``|``ParameterInvalidJsonValueError``

		---

		### Parameters Helper

		Parameters are built into SQL at the application level; that is, the SQL and separate parameters are not sent to the database;
		although most databases support this type of operation, the PySQLXEngine does it before calling the database to avoid possible incompatibilities.
		This allows you to show the precompiled queries and send only raw SQL while maintaining minimal consistency across types.

		#### SQL with parameters syntax
		    * SQL: ``SELECT * FROM table WHERE id = :id``
		    * Parameters: ``{"id": 1}``

		#### Parameters(dict):

		    * dict ``key`` must be a valid string.
		    * dict ``value`` can be one of the types: (
		        ``bool``,
		        ``bytes``,
		        ``date``,
		        ``datetime``,
		        ``Decimal``,
		        ``dict``,
		        ``Enum``, # Enum must be a subclass of enum.Enum
		        ``float``,
		        ``int``,
		        ``list``,
		        ``str``,
		        ``time``,
		        ``tuple``,
		        ``UUID``,
		        ``None``
		    )

		#### Python types vs SQL types:

		    [Documentation](https://carlos-rian.github.io/pysqlx-engine/type_mappings/)

		```
		    * bool     -> bool|bit|boolean|tinyint|etc
		    * bytes    -> bytea|binary|varbinary|etc
		    * date     -> date|nvarchar|varchar|string|etc
		    * datetime -> timestamp|timestamptz|datetime|datetime2|nvarchar|varchar|string|etc
		    * Decimal  -> decimal|numeric|etc
		    * dict     -> json|jsonb|nvarchar|varchar|string|etc
		    * float    -> float|real|numeric|etc
		    * int      -> int|integer|smallint|bigint|tinyint|etc
		    * list     -> json|jsonb|nvarchar|varchar|string|etc
		    * str      -> varchar|text|nvarchar|char|etc
		    * time     -> time|nvarchar|varchar|string|etc
		    * tuple    -> array(Postgres Native), another database: error.
		    * UUID     -> uuid|varchar|text|nvarchar|etc
		    * Enum     -> varchar|text|nvarchar|etc
		    * None     -> null
		```

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    result = db.query_first_as_dict(sql="SELECT 1 as id, 'rian' as name")
		    print(result)
		    # output -> {'id': 1, 'name': 'rian'}

		    result = db.query_first_as_dict(sql="SELECT 1 as :id, 'rian' as name", parameters={"id": 1})
		    print(result)
		    # output -> {'id': 1, 'name': 'rian'}

		    db.close()
		```
		"""
		...
	# --
	def execute(self, sql: str, parameters: Optional[DictParam] = None) -> "int":
		"""
		## Description

		Executes a query/sql and returns the number of rows affected.

		---

		### Helper

		    * Arguments:

		        ``sql(str)``:  sql to be executed.

		        ``parameters(dict)``: (Default is None) parameters must be a dictionary with the name of the parameter and the value.

		    * Returns: ``int``: number of rows affected.

		    * Raises: ``ExecuteError``|``TypeError`` | ``ParameterInvalidProviderError``|``ParameterInvalidValueError``|``ParameterInvalidJsonValueError``

		### Parameters Helper

		Parameters are built into SQL at the application level; that is, the SQL and separate parameters are not sent to the database;
		although most databases support this type of operation, the PySQLXEngine does it before calling the database to avoid possible incompatibilities.
		This allows you to show the precompiled queries and send only raw SQL while maintaining minimal consistency across types.

		#### SQL with parameters syntax
		    * SQL: ``SELECT * FROM table WHERE id = :id``
		    * Parameters: ``{"id": 1}``

		#### Parameters(dict):

		    * dict ``key`` must be a valid string.
		    * dict ``value`` can be one of the types: (
		        ``bool``,
		        ``bytes``,
		        ``date``,
		        ``datetime``,
		        ``Decimal``,
		        ``dict``,
		        ``Enum``, # Enum must be a subclass of enum.Enum
		        ``float``,
		        ``int``,
		        ``list``,
		        ``str``,
		        ``time``,
		        ``tuple``,
		        ``UUID``,
		        ``None``
		    )

		#### Python types vs SQL types:

		    [Documentation](https://carlos-rian.github.io/pysqlx-engine/type_mappings/)

		```
		    * bool     -> bool|bit|boolean|tinyint|etc
		    * bytes    -> bytea|binary|varbinary|etc
		    * date     -> date|nvarchar|varchar|string|etc
		    * datetime -> timestamp|timestamptz|datetime|datetime2|nvarchar|varchar|string|etc
		    * Decimal  -> decimal|numeric|etc
		    * dict     -> json|jsonb|nvarchar|varchar|string|etc
		    * float    -> float|real|numeric|etc
		    * int      -> int|integer|smallint|bigint|tinyint|etc
		    * list     -> json|jsonb|nvarchar|varchar|string|etc
		    * str      -> varchar|text|nvarchar|char|etc
		    * time     -> time|nvarchar|varchar|string|etc
		    * tuple    -> array(Postgres Native), another database: error.
		    * UUID     -> uuid|varchar|text|nvarchar|etc
		    * Enum     -> varchar|text|nvarchar|etc
		    * None     -> null
		```

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    result = db.execute("INSERT INTO users (name) VALUES ('rian')")
		    print(f"rows_affected = {result}")
		    # output -> rows_affected = 1
		```
		"""
		...
	def set_isolation_level(self, isolation_level: ISOLATION_LEVEL) -> "None":
		"""
		## Description

		Sets the isolation level of the connection.

		The isolation level is set before the transaction is started.
		Is used to separate the transaction per level.

		The `Snapshot` isolation level is supported by MS SQL Server.

		The Sqlite does not support the isolation level.

		---

		### Helper

		    * Arguments: ``isolation_level(str)``: isolation level to be set (
		        ReadUncommitted,
		        ReadCommitted,
		        RepeatableRead,
		        Snapshot,
		        Serializable
		    )

		    * Returns: ``None``

		    * Raises: ``IsolationLevelError``, ``ValueError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()
		    db.set_isolation_level(isolation_level="ReadUncommitted")
		```
		---

		### Isolation Level Help
		    * [MSSQL](https://learn.microsoft.com/en-us/sql/t-sql/language-elements/transaction-isolation-levels)
		    * [Postgres](https://www.postgresql.org/docs/current/sql-set-transaction.html)
		    * [MySQL](https://dev.mysql.com/doc/refman/8.0/en/innodb-transaction-isolation-levels.html)
		    * [SQLite](https://www.sqlite.org/isolation.html)
		"""
		...
	def begin(self) -> "None":
		"""
		## Description

		Starts a transaction using ``BEGIN``.

		``begin()`` is equivalent to `start_transaction()` without setting the isolation level.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``None``

		    * Raises: ``RawCmdError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()
		    db.begin()
		```
		"""
		...
	def commit(self) -> "None":
		"""
		## Description

		Commits the current transaction.

		The `begin()` method must be called before calling `commit()`.

		If the database not need set the isolation level, maybe you can not use `begin()` and `commit()`.

		The PySQLXEngine by default uses the `begin()` and `commit()` in all transactions.

		---

		### Helper

		    * Arguments: ``None`

		    * Returns: ``None``

		    * Raises: ``RawCmdError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    db.begin()
		    db.execute("CREATE TABLE users (id serial PRIMARY KEY, name varchar(255))")
		    db.execute("INSERT INTO users (name) VALUES ('rian')")
		    db.commit()
		```
		"""
		...
	def rollback(self) -> "None":
		"""
		## Description

		Rollbacks the current transaction.

		Rollback is used to cancel the transaction, when you uses the rollback,
		the transaction is canceled and the changes are not saved.

		The ``begin()`` method must be called before calling ``rollback()``.

		If the database not need set the isolation level, maybe you can not use ``begin()`` and ``rollback()``.

		The PySQLXEngine by default try uses the ``begin()`` and ``commit()`` in all transactions.

		---

		### Helper

		    * Arguments: ``None``

		    * Returns: ``None``

		    * Raises: ``RawCmdError``

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    db.begin()
		    db.execute("CREATE TABLE users (id serial PRIMARY KEY, name varchar(255))")
		    db.execute("INSERT INTO users (name) VALUES ('rian')")
		    db.rollback()
		```
		"""
		...
	def start_transaction(self, isolation_level: Union[ISOLATION_LEVEL, None] = None) -> "None":
		"""
		## Description

		Starts a transaction with ``BEGIN/BEGIN TRANSACTION``. by default, does not set the isolation level.

		The ``Snapshot`` isolation level is supported by MS SQL Server.

		The Sqlite does not support the isolation level.

		---

		### Helper

		    * Arguments: ``isolation_level(str)``: by default is None. Isolation level to be set (
		        ReadUncommitted,
		        ReadCommitted,
		        RepeatableRead,
		        Snapshot,
		        Serializable
		    )

		    * Returns: `None`

		    * Raises: (`IsolationLevelError`, `StartTransactionError` `ValueError`)

		---

		### Example
		```python
		    from pysqlx_engine import PySQLXEngineSync

		    uri = "postgresql://user:pass@host:port/db?schema=sample"
		    db = PySQLXEngineSync(uri=uri)
		    db.connect()

		    # with isolation level
		    db.start_transaction(isolation_level="ReadCommitted")

		    # without isolation level
		    db.start_transaction()
		```

		---

		### Isolation Level Help
		    * [MSSQL](https://learn.microsoft.com/en-us/sql/t-sql/language-elements/transaction-isolation-levels)
		    * [Postgres](https://www.postgresql.org/docs/current/sql-set-transaction.html)
		    * [MySQL](https://dev.mysql.com/doc/refman/8.0/en/innodb-transaction-isolation-levels.html)
		    * [SQLite](https://www.sqlite.org/isolation.html)
		"""
		...



from types import TracebackType
from typing import Optional, Type
from .conn import PySQLXEngineSync

class PySQLXEnginePoolSync:
	"""
	The PySQLXEnginePoolSync class is a simple connection pool for PySQLXEngine.
	It is thread-safe and can be used in thread or process to run multiple queries in parallel.

	Example:
	```python
	    import asyncio
	    from pysqlx_engine import PySQLXEnginePoolSync

	    pool = PySQLXEnginePoolSync("postgresql://user:pass@host:port/db?schema=sample", max_connections=3)

	    def main():
	        db1 = pool.new_connection()
	        db2 = pool.new_connection()
	        db3 = pool.new_connection()

	        # multiple queries can be run in parallel using threads or processes
	        results = [
	                db1.fetch("SELECT 1 AS one"),
	                db2.fetch("SELECT 2 AS two"),
	                db3.fetch("SELECT 3 AS three")
	            ]
	        print(results)
	        # [[BaseRow(one=1)], [BaseRow(two=2)], [BaseRow(three=3)]]

	        # try creating a connection when the pool is full
	        db4 = pool.new_connection() # raises PoolMaxConnectionsError

	    main()
	```
	"""

	uri: str
	max_connections: int

	def __init__(self, uri: str, max_connections: Optional[int] = None): ...
	def __enter__(self) -> "PySQLXEngineSync":
		"""
		## Description
		Create a pool instance. using `with`
		"""
		...
	def __exit__(
		self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], exc_tb: Optional[TracebackType]
	): ...
	def new_connection(self) -> PySQLXEngineSync:
		"""
		## Description
		Create a new connection from the pool.
		this method tries to get a connection from the pool, if the pool is full to this instance, it will raise `PoolMaxConnectionsError`.

		If max_connections is None, the pool is "unlimited," stopped only by the database server's connection limit.

		* Arguments: `None`:

		* Returns: `PySQLXEngine`

		* Raises: `PoolMaxConnectionsError`, `ConnectError`

		## Example
		```python
		from pysqlx_engine import PySQLXEnginePoolSync

		pool = PySQLXEnginePoolSync("postgresql://user:pass@host:port/db?schema=sample", max_connections=3)

		def main():
		    db1 = pool.new_connection()  # return a new PySQLXEngineSync instance
		    db2 = pool.new_connection()  # return a new PySQLXEngineSync instance
		    db3 = pool.new_connection()  # return a new PySQLXEngineSync instance
		```
		"""
		...



from datetime import date, datetime, time
from decimal import Decimal
from os import getenv
from typing import Union
from uuid import UUID

from pydantic import VERSION as PYDANTIC_VERSION
from pydantic import BaseModel as BaseConfig
from pydantic.types import Json
from typing_extensions import Literal

PYDANTIC_IS_V1 = PYDANTIC_VERSION < "2.0.0"


TYPES_OUT = {
	"bool": bool,
	"str": str,
	"int": int,
	"list": tuple,
	"json": Union[dict, list, Json],
	"uuid": UUID,
	"time": time,
	"date": date,
	"datetime": datetime,
	"float": float,
	"bytes": bytes,
	"decimal": Decimal,
}

TYPES_IN = TYPES_OUT.copy()
TYPES_IN.update({"array": list, "dict": dict})
TYPES_IN.pop("json")

ISOLATION_LEVEL = Literal["ReadUncommitted", "ReadCommitted", "RepeatableRead", "Snapshot", "Serializable"]

PROVIDER = Literal["postgresql", "mysql", "sqlserver", "sqlite"]


CODE_AlreadyConnectedError = "PYSQLX001"
CODE_PoolMaxConnectionsError = "PYSQLX002"
CODE_ParameterInvalidProviderError = "PYSQLX003"
CODE_ParameterInvalidValueError = "PYSQLX004"
CODE_ParameterInvalidJsonValueError = "PYSQLX005"


class LogConfig(BaseConfig):
	PYSQLX_DEV_MODE: bool = getenv("PYSQLX_DEV_MODE", "0") != "0"
	PYSQLX_SQL_LOG: bool = getenv("PYSQLX_SQL_LOG", "0") != "0"
	PYSQLX_USE_COLOR: bool = getenv("PYSQLX_USE_COLOR", "0") != "0"
	PYSQLX_ERROR_JSON_FMT: bool = getenv("PYSQLX_ERROR_JSON_FMT", "0") != "0"


LOG_CONFIG = LogConfig()
"""
## Description

CONFIG constant for PySQLXEngine, is used to configure the log and exception messages.

You can set the following ``environment variables``:
    - `PYSQLX_DEV_MODE`
    - `PYSQLX_SQL_LOG`
    - `PYSQLX_USE_COLOR`
    - `PYSQLX_ERROR_JSON_FMT`

---

### Helper
    * `PYSQLX_DEV_MODE`: bool = False
        If True, the development mode will be activated, showing the sql builded before executing.

    * `PYSQLX_SQL_LOG`: bool = False
        If True, the SQL statements will be printed in the console.
    * `PYSQLX_USE_COLOR`: bool = False
        If True, the messages will be printed in color.
    * `PYSQLX_ERROR_JSON_FMT`: bool = False
        If True, the error messages will be printed in JSON format.

Or you can set the value of the variables in the code.

---

### Example:
```python
    # config the pysqlx_engine log
    from pysqlx_engine import LOG_CONFIG

    LOG_CONFIG.PYSQLX_SQL_LOG = True
    LOG_CONFIG.PYSQLX_USE_COLOR = True
    LOG_CONFIG.PYSQLX_ERROR_JSON_FMT = True

    # set the log info level when the PYSQLX_SQL_LOG is True
    import logging
    logging.basicConfig(level=logging.INFO)

    # connect to the database
    from pysqlx_engine import PySQLXEngine

    db = PySQLXEngine(uri="postgresql://user:pass@host:port/db?schema=sample")
    db.connect()

    # run the query and see the log
    db.query("SELECT * FROM table")
    # output -> INFO:root:SELECT * FROM table

    # run the invalid query and see the error in JSON format
    db.query("SELECT * FROM invalid_table")
    # output ->  INFO:root:SELECT * FROM invalid_table
    
    # raise the error
    pysqlx_engine._core.errors.QueryError: 
    {
       "code": "42P01",
       "message": "relation 'invalid_table' does not exist",
       "error": "QueryError"
    }
```

"""



import asyncio
import functools
import shutil
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from functools import lru_cache
from typing import Callable, List, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel
from pysqlx_core import PySQLxError as _PySQLXError
from typing_extensions import ParamSpec

from .const import ISOLATION_LEVEL, PYDANTIC_IS_V1
from .errors import (
	ConnectError,
	ExecuteError,
	IsoLevelError,
	NotConnectedError,
	PySQLXError,
	QueryError,
	RawCmdError,
	StartTransactionError,
)
from .helper import isolation_error_message, parameters_type_error_message, sql_type_error_message
from .param_converter import convert

if PYDANTIC_IS_V1:
	from pydantic import parse_obj_as as _parse_obj_as  # pragma: no cover
else:
	from pydantic import TypeAdapter  # pragma: no cover


P = ParamSpec("P")
T = TypeVar("T")


def force_sync(fn: Callable[P, T]) -> Callable[P, T]:
	"""
	turn an async function to sync function
	"""

	@functools.wraps(fn)
	def wrapper(*args: P.args, **kwds: P.kwargs) -> T:
		res = fn(*args, **kwds)
		loop = asyncio.get_event_loop()

		# if loop is already running, we don't need to run it again
		if loop.is_running():
			return res
		# if the result is a coroutine, we need to run it in the loop synchronously
		elif asyncio.iscoroutine(res):
			return loop.run_until_complete(res)

		# case the result is not a coroutine, we just return it
		return res

	return wrapper


def pysqlx_get_error(err: _PySQLXError) -> PySQLXError:
	types_err = {
		"ConnectError": ConnectError,
		"ExecuteError": ExecuteError,
		"NotConnectedError": NotConnectedError,
		"IsoLevelError": IsoLevelError,
		"PySQLXError": PySQLXError,
		"QueryError": QueryError,
		"RawCmdError": RawCmdError,
		"StartTransactionError": StartTransactionError,
	}
	error: PySQLXError = types_err.get(err.error())
	if error is not None:
		return error(err=err)
	return err


def check_sql_and_parameters(sql: str, parameters: dict):
	if not isinstance(sql, str):
		raise TypeError(sql_type_error_message())

	if parameters is not None:
		_types = (bool, str, int, list, dict, tuple, UUID, time, date, datetime, float, bytes, Decimal, Enum)
		if (
			not isinstance(parameters, dict)
			or not all([isinstance(key, str) for key in parameters.keys()])
			or not all([isinstance(value, _types) or value is None for value in parameters.values()])
		):
			raise TypeError(parameters_type_error_message())


@lru_cache(maxsize=None)
def check_isolation_level(isolation_level: ISOLATION_LEVEL):
	levels = [
		"ReadUncommitted",
		"ReadCommitted",
		"RepeatableRead",
		"Snapshot",
		"Serializable",
	]
	if not isinstance(isolation_level, str) or not any([isolation_level == level for level in levels]):
		raise ValueError(isolation_error_message())


def build_sql(provider: str, sql: str, parameters: dict = None) -> str:
	new_sql = sql
	if parameters is not None:
		load_parameter = {key: convert(provider=provider, value=value, field=key) for key, value in parameters.items()}

		param_as_list_of_tuples = sorted(load_parameter.items(), key=lambda x: x[0], reverse=True)

		for key, value in param_as_list_of_tuples:
			new_sql = new_sql.replace(f":{key}", str(value))
	return new_sql


def parse_obj_as(type_: Union[BaseModel, List[BaseModel]], obj: Union[dict, list]) -> type:
	return _parse_obj_as(type_=type_, obj=obj) if PYDANTIC_IS_V1 else TypeAdapter(type=type_).validate_python(obj)


def create_log_line(text: str, character: str = "=") -> str:
	# Get the terminal width
	terminal_width = shutil.get_terminal_size().columns

	# If the text is longer than the terminal width, truncate it
	if len(text) > terminal_width:
		text = text[:terminal_width]

	# Calculate the space needed to center the text
	space = (terminal_width - len(text)) // 2

	# Create the line
	line = f"{character * space}{text}{character * space}"

	# If there's any difference, adjust the line to match the terminal width
	if len(line) < terminal_width:
		line += character * (terminal_width - len(line))  # pragma: no cover

	return line



from enum import Enum
from typing import Union

from .const import PROVIDER
from .errors import ParameterInvalidValueError
from .param import SupportedValueType, get_method, try_enum, try_tuple_enum


def convert(provider: PROVIDER, value: SupportedValueType, field: str = "") -> Union[str, int, float]:
	if value is None:
		return "NULL"

	elif isinstance(value, Enum):
		return try_enum(provider, value, field)

	elif isinstance(value, tuple) and len(value) > 0 and isinstance(value[0], Enum):
		return try_tuple_enum(provider, value, field)

	typ_ = type(value)
	method = get_method(typ=typ_)
	if method is None:
		raise ParameterInvalidValueError(
			field=field,
			provider=provider,
			typ_from=typ_,
			typ_to="str|int|float|etc",
			details="invalid type, the value is not a allowed type.",
		)

	return method(provider, value, field)



from types import TracebackType
from typing import Optional, Type
from .aconn import PySQLXEngine

class PySQLXEnginePool:
	"""
	The PySQLXEnginePool class is a simple connection pool for PySQLXEngine.
	It is thread-safe and can be used in async code to run multiple queries in parallel.

	Example:
	```python
	    import asyncio
	    from pysqlx_engine import PySQLXEnginePool

	    pool = PySQLXEnginePool("postgresql://user:pass@host:port/db?schema=sample", max_connections=3)

	    async def main():
	        db1 = await pool.new_connection()
	        db2 = await pool.new_connection()
	        db3 = await pool.new_connection()

	        # multiple queries can be run in parallel
	        coro = [
	                await db1.query(sql="SELECT 1 AS one"),
	                await db2.query(sql="SELECT 2 AS two"),
	                await db3.query(sql="SELECT 3 AS three")
	            ]
	        # wait for all queries to finish
	        results = await asyncio.gather(*coro)
	        print(results)
	        # [[BaseRow(one=1)], [BaseRow(two=2)], [BaseRow(three=3)]]

	        # try creating a connection when the pool is full
	        db4 = await pool.new_connection() # raises PoolMaxConnectionsError

	    asyncio.run(main())
	```
	"""

	uri: str
	max_connections: int

	def __init__(self, uri: str, max_connections: Optional[int] = None): ...
	async def __aenter__(self) -> "PySQLXEngine":
		"""
		## Description
		Create a pool instance. using `async with`
		"""
		...
	async def __aexit__(
		self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], exc_tb: Optional[TracebackType]
	): ...
	async def new_connection(self) -> PySQLXEngine:
		"""
		## Description
		Create a new connection from the pool.
		this method tries to get a connection from the pool, if the pool is full to this instance,
		it will raise `PoolMaxConnectionsError`.

		If max_connections is None, the pool is "unlimited," stopped only by the database server's connection limit.

		* Arguments: `None`:

		* Returns: `PySQLXEngine`

		* Raises: `PoolMaxConnectionsError`, `ConnectError`

		## Example
		```python
		from pysqlx_engine import PySQLXEnginePool

		pool = PySQLXEnginePool("postgresql://user:pass@host:port/db?schema=sample", max_connections=3)

		async def main():
		    db1 = await pool.new_connection()  # return a new PySQLXEngine instance
		    db2 = await pool.new_connection()  # return a new PySQLXEngine instance
		    db3 = await pool.new_connection()  # return a new PySQLXEngine instance
		```
		"""
		...



from ._core.errors import AlreadyConnectedError as AlreadyConnectedError
from ._core.errors import ConnectError as ConnectError
from ._core.errors import ExecuteError as ExecuteError
from ._core.errors import IsoLevelError as IsoLevelError
from ._core.errors import NotConnectedError as NotConnectedError
from ._core.errors import ParameterInvalidJsonValueError as ParameterInvalidJsonValueError
from ._core.errors import ParameterInvalidProviderError as ParameterInvalidProviderError
from ._core.errors import ParameterInvalidValueError as ParameterInvalidValueError
from ._core.errors import PoolMaxConnectionsError as PoolMaxConnectionsError
from ._core.errors import PySQLXError as PySQLXError
from ._core.errors import QueryError as QueryError
from ._core.errors import RawCmdError as RawCmdError
from ._core.errors import StartTransactionError as StartTransactionError



from ._core.aconn import PySQLXEngine as PySQLXEngine
from ._core.apool import PySQLXEnginePool as PySQLXEnginePool
from ._core.pool import PySQLXEnginePoolSync as PySQLXEnginePoolSync
from ._core.conn import PySQLXEngineSync as PySQLXEngineSync
from ._core.parser import BaseRow as BaseRow

from ._core.const import LOG_CONFIG as LOG_CONFIG



from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, Dict, List, Sequence, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, create_model
from pysqlx_core import PySQLxResponse

from .const import PYDANTIC_IS_V1, TYPES_OUT
from .util import build_sql, parse_obj_as

MyModel = TypeVar("MyModel", bound="BaseRow")

JsonParam = Union[
	Dict[str, Union[bool, str, int, UUID, time, date, datetime, float, bytes, Decimal, None]],
	List[Dict[str, Union[bool, str, int, UUID, time, date, datetime, float, bytes, Decimal, None]]],
	Dict,
	List,
]
DictParam = Dict[
	str,
	Union[bool, str, int, JsonParam, UUID, time, date, datetime, float, bytes, Decimal, None],
]


class BaseRow(BaseModel):
	"""
	BaseRow class for a row returned by PySQLXEngine.

	BaseRow is a class that represents a row of the result of a query.

	BaseRow is a class created from `Pydantic`, then you have all the benefits of `Pydantic`.
	"""

	def get_columns(self) -> Dict[str, Any]:
		"""
		Return the columns of the row.

		Returns:
		    Dict[str, Any]: The columns of the row.
		"""
		values = self.dict() if PYDANTIC_IS_V1 else self.model_dump()
		columns = {}
		for key, value in values.items():
			columns[key] = type(value)
		return columns

	def __str__(self) -> str:
		return super().__repr__()


class ParserIn:
	__slots__ = ("result", "model")

	def __init__(self, result: PySQLxResponse, model: MyModel = None):
		self.result: PySQLxResponse = result
		self.model: MyModel = model

	def create_model(self) -> BaseRow:
		fields = {}
		for key, value in self.result.get_types().items():
			if value.startswith("array_"):
				_, v = value.split("_")
				type_ = TYPES_OUT.get(v, Any)
				fields[key] = (Union[Sequence[type_], None], None)
			else:
				type_ = TYPES_OUT.get(value, Any)
				fields[key] = (Union[type_, None], None)

		model = create_model("BaseRow", **fields, __base__=BaseRow)
		return model

	def parse(self) -> List[BaseRow]:
		if len(self.result) == 0:
			return []
		model = self.model or self.create_model()
		return parse_obj_as(type_=List[model], obj=self.result.get_all())

	def parse_first(self) -> Union[BaseRow, None]:
		if len(self.result) == 0:
			return None
		model = self.model or self.create_model()
		return parse_obj_as(type_=model, obj=self.result.get_first())


class ParserSQL:
	def __init__(self, provider: str, sql: str, parameters: dict = None) -> None:
		self._sql: str = sql
		self.provider: str = provider
		self.parameters: dict = parameters

	def sql(self) -> str:
		return build_sql(provider=self.provider, sql=self._sql, parameters=self.parameters)



"""
This module contains the errors that can be raised by the PySQLXEngine.

* PySQLXError
* QueryError
* ExecuteError
* ConnectError
* IsoLevelError
* StartTransactionError
* NotConnectedError

"""

from typing import Any

from pysqlx_core import PySQLxError as _PySQLXError

from .const import (
	LOG_CONFIG,
	PROVIDER,
	CODE_AlreadyConnectedError,
	CODE_ParameterInvalidJsonValueError,
	CODE_ParameterInvalidProviderError,
	CODE_ParameterInvalidValueError,
	CODE_PoolMaxConnectionsError,
)
from .helper import fe_json


class PySQLXError(Exception):
	"""Base class for all PySQLXEngine errors."""

	def __init__(self, err: _PySQLXError, *args: object):
		self.code: str = err.code()
		self.message: str = err.message()
		self._type: str = err.error()

		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": self.code,
					"message": self.message,
					"error": self._type,
				}
			)
			super().__init__(msg)
		else:
			msg = f"{self._type}(code='{self.code}', message='{self.message}')"
			super().__init__(msg)

	def error(self) -> str:
		"""Return the error type."""
		return self._type


class QueryError(PySQLXError):
	"""
	Raised when a query fails.
	"""

	...


class ExecuteError(PySQLXError):
	"""
	Raised when an error occurs while executing a query or invalid sql.
	"""

	...


class ConnectError(PySQLXError):
	"""Raised when a connection error occurs."""

	...


class IsoLevelError(PySQLXError):
	"""
	Raised when the isolation level is not supported or not valid
	"""

	...


class StartTransactionError(PySQLXError):
	"""
	Raised when the user tries to start a transaction
	while there is already an active transaction for example.
	"""

	...


class RawCmdError(PySQLXError):
	"""
	Raised when the user tries to execute a raw command

	"""

	...


class NotConnectedError(Exception):
	"""
	Raised when the user tries to execute a sql but is not connected to the database.
	"""

	...


class AlreadyConnectedError(Exception):
	"""
	Raised when the user tries to connect to the database but is already connected.
	"""

	def __init__(self, *args: object) -> None:
		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": CODE_AlreadyConnectedError,
					"message": "Already connected to the database",
					"error": "AlreadyConnectedError",
				}
			)

			super().__init__(msg)
		else:
			msg = (
				f"AlreadyConnectedError(code='{CODE_AlreadyConnectedError}', "
				"message='Already connected to the database')"
			)
			super().__init__(msg)


class PoolMaxConnectionsError(Exception):
	"""
	Raised when the user tries to get a connection from the pool
	but the maximum number of connections has been reached.
	"""

	def __init__(self, *args: object) -> None:
		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": CODE_PoolMaxConnectionsError,
					"message": "maximum number of connections reached",
					"error": "PoolMaxConnectionsError",
				}
			)

			super().__init__(msg)
		else:
			msg = (
				f"PoolMaxConnectionsError(code='{CODE_PoolMaxConnectionsError}', "
				"message='maximum number of connections reached')"
			)
			super().__init__(msg)


class ParameterInvalidProviderError(Exception):
	"""
	Raised when the user tries to pass an invalid type to a provider.
	"""

	def __init__(self, field: str, provider: PROVIDER, typ: Any) -> None:
		self.field = field
		self.provider = provider
		self.type = typ

		message = f"the provider '{self.provider}' does not support the type '{self.type}'."

		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": CODE_ParameterInvalidProviderError,
					"message": message,
					"error": "ParameterInvalidError",
				}
			)
			super().__init__(msg)
		else:
			msg = f"ParameterInvalidError(code='{CODE_ParameterInvalidProviderError}', " f"message='{message}')"
			super().__init__(msg)


class ParameterInvalidValueError(Exception):
	"""
	Raised when the user tries to pass an invalid parameter to the sql.
	"""

	def __init__(self, field: str, provider: PROVIDER, typ_from: Any, typ_to: Any, details: str = None) -> None:
		self.field = field
		self.provider = provider
		self.typ_from = typ_from
		self.typ_to = typ_to
		self.details = details

		message = (
			f"pysqlx-engine for the provider: {self.provider} had an error converting the type "
			f"'{self.typ_from}' to type '{self.typ_to}'. types supported: "
			"(bool, bytes, date, datetime, Decimal, dict, float, int, list, str, time, tuple, UUID, None)."
		)

		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": CODE_ParameterInvalidValueError,
					"message": message,
					"error": "ParameterInvalidError",
					"details": details,
				}
			)
			super().__init__(msg)
		else:
			msg = (
				f"ParameterInvalidError(code='{CODE_ParameterInvalidValueError}', "
				f"message='{message}', details='{details}')"
			)
			super().__init__(msg)


class ParameterInvalidJsonValueError(Exception):
	"""
	Raised when the user tries to pass an invalid value parameter to convert json.
	"""

	def __init__(self, typ_from: str, typ_to: str, details: str = None) -> None:
		self.typ_from = typ_from
		self.typ_to = typ_to
		self.details = details

		message = (
			f"error to convert the type '{self.typ_from}' to json value. types supported: "
			"(bool, bytes, date, datetime, Decimal, dict, float, int, list, str, time, tuple, UUID, None)."
		)

		if LOG_CONFIG.PYSQLX_ERROR_JSON_FMT:
			msg = fe_json(
				{
					"code": CODE_ParameterInvalidJsonValueError,
					"message": message,
					"error": "ParameterInvalidError",
					"details": details,
				}
			)
			super().__init__(msg)
		else:
			msg = (
				f"ParameterInvalidError(code='{CODE_ParameterInvalidJsonValueError}', "
				f"message='{message}', details='{self.details}')"
			)
			super().__init__(msg)



from typing import Optional

import pysqlx_core

from .._core.const import LOG_CONFIG
from .._core.logger import logger
from .._core.parser import BaseRow, MyModel, ParserIn, ParserSQL

# ParserSQL,
# import necessary using _core to not subscribe default parser
from .const import ISOLATION_LEVEL
from .errors import AlreadyConnectedError, NotConnectedError, ParameterInvalidValueError
from .helper import fe_sql, model_parameter_error_message, not_connected_error_message
from .util import check_isolation_level, check_sql_and_parameters, create_log_line, pysqlx_get_error


class PySQLXEngine:
	__slots__ = ["uri", "connected", "_conn", "_provider"]

	uri: str
	connected: bool

	def __init__(self, uri: str):
		self.connected: bool = False

		_providers = ["postgresql", "mysql", "sqlserver", "sqlite"]
		if not uri or not any([uri.startswith(prov) for prov in [*_providers, "file"]]):
			raise ValueError(f"invalid uri: {uri}, check the usage uri, accepted providers: {', '.join(_providers)}")

		if uri.startswith("sqlite"):
			uri = uri.replace("sqlite", "file", 1)

		self.uri: str = uri
		self._conn: Optional[pysqlx_core.Connection] = None

		self._provider = "sqlite"

		for prov in _providers:
			if self.uri.startswith(prov):
				self._provider = prov

	def __del__(self):
		if self.connected:
			del self._conn
			self._conn = None
			self.connected = False

	def is_healthy(self):
		self._pre_validate()
		return self._conn.is_healthy()

	def requires_isolation_first(self):
		self._pre_validate()
		return self._conn.requires_isolation_first()

	async def __aenter__(self):
		await self.connect()
		return self

	async def __aexit__(self, exc_type, exc, exc_tb):
		await self.close()

	async def connect(self):
		if self.connected:
			raise AlreadyConnectedError()
		try:
			self._conn = await pysqlx_core.new(uri=self.uri)
			self.connected = True
		except pysqlx_core.PySQLxError as e:
			raise pysqlx_get_error(err=e)

	async def close(self):
		if self.connected:
			del self._conn
			self._conn = None
			self.connected = False

	async def raw_cmd(self, sql: str):
		return await self._run(func=self._conn.raw_cmd, sql=sql)

	def _dev_mode(self, sql: str, parameters: Optional[dict] = None):
		logger.debug(create_log_line(" PYSQLX-ENGINE DEVELOPMENT MODE "))
		logger.debug(
			create_log_line(
				"THIS SQL STATEMENT IS PRE BUILD JUST FOR DEVELOPMENT PURPOSE, "
				"ALL SQL AND PARAMS WILL PASS TO DATABASE."
			)
		)
		log_sql = ParserSQL(provider=self._provider, sql=sql, parameters=parameters).sql()
		logger.debug(fe_sql(sql=log_sql))

	async def _run(self, func, sql: str, parameters: Optional[dict] = None, model: Optional[MyModel] = None):
		if model is not None and not issubclass(model, BaseRow):
			raise TypeError(model_parameter_error_message())

		try:
			stmt = pysqlx_core.PySQLxStatement(provider=self._provider, sql=sql, params=parameters)
			if LOG_CONFIG.PYSQLX_DEV_MODE:
				self._dev_mode(sql=sql, parameters=parameters)

			if LOG_CONFIG.PYSQLX_SQL_LOG:
				logger.debug(create_log_line(" THIS SQL IS THE FINAL SQL AND PARAMS STATEMENT THAT WILL BE EXECUTED "))
				logger.info(f"\nSQL: {fe_sql(stmt.sql())}\nPARAMS: {stmt.params()}")

			return await func(stmt)

		except pysqlx_core.PySQLxError as e:
			raise pysqlx_get_error(err=e)

		except pysqlx_core.PySQLxInvalidParamError as e:
			raise ParameterInvalidValueError(
				field=e.field(),
				provider=self._provider,
				typ_from=e.typ_from(),
				typ_to=e.typ_to(),
			)

	async def query(self, sql: str, parameters: Optional[dict] = None, model: Optional[MyModel] = None):
		self._pre_validate(sql=sql, parameters=parameters)
		result = await self._run(func=self._conn.query_typed, sql=sql, parameters=parameters, model=model)
		return ParserIn(result=result, model=model).parse()

	async def query_as_dict(self, sql: str, parameters: Optional[dict] = None):
		self._pre_validate(sql=sql, parameters=parameters)
		return await self._run(self._conn.query_all, sql=sql, parameters=parameters)

	async def query_first(self, sql: str, parameters: Optional[dict] = None, model: Optional[MyModel] = None):
		self._pre_validate(sql=sql, parameters=parameters)
		result = await self._run(self._conn.query_typed, sql=sql, parameters=parameters, model=model)
		return ParserIn(result=result, model=model).parse_first()

	async def query_first_as_dict(self, sql: str, parameters: Optional[dict] = None):
		self._pre_validate(sql=sql, parameters=parameters)
		row = await self._run(self._conn.query_one, sql=sql, parameters=parameters)
		return row if row else None

	async def execute(self, sql: str, parameters: Optional[dict] = None):
		self._pre_validate(sql=sql, parameters=parameters)
		return await self._run(self._conn.execute, sql=sql, parameters=parameters)

	async def set_isolation_level(self, isolation_level: ISOLATION_LEVEL):
		self._pre_validate(isolation_level=isolation_level)
		try:
			await self._conn.set_isolation_level(isolation_level=isolation_level)
		except pysqlx_core.PySQLxError as e:
			raise pysqlx_get_error(err=e)

	async def begin(self):
		await self.start_transaction()

	async def commit(self):
		self._pre_validate()
		if self._provider == "sqlserver":
			await self.raw_cmd(sql="COMMIT TRANSACTION;")
		else:
			await self.raw_cmd(sql="COMMIT;")

	async def rollback(self):
		self._pre_validate()
		if self._provider == "sqlserver":
			await self.raw_cmd(sql="ROLLBACK TRANSACTION;")
		else:
			await self.raw_cmd(sql="ROLLBACK;")

	async def start_transaction(self, isolation_level: Optional[ISOLATION_LEVEL] = None):
		self._pre_validate(isolation_level=isolation_level)
		try:
			await self._conn.start_transaction(isolation_level=isolation_level)
		except pysqlx_core.PySQLxError as e:
			raise pysqlx_get_error(err=e)

	def _pre_validate(
		self,
		sql: str = "",
		parameters: Optional[dict] = None,
		isolation_level: Optional[ISOLATION_LEVEL] = None,
	):
		if not self.connected:
			raise NotConnectedError(not_connected_error_message())

		if sql != "":
			check_sql_and_parameters(sql=sql, parameters=parameters)

		if isolation_level is not None:
			check_isolation_level(isolation_level=isolation_level)



import json
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from functools import lru_cache
from typing import Any, Callable, Dict, List, Tuple, Type, Union
from uuid import UUID

from pysqlx_engine._core.json_econder import PySQLXJsonEnconder

from .const import PROVIDER
from .errors import ParameterInvalidProviderError, ParameterInvalidValueError

SupportedValueType = Union[
	bool,
	str,
	int,
	Dict[str, Any],
	List[Dict[str, Any]],
	UUID,
	time,
	date,
	datetime,
	float,
	bytes,
	Decimal,
	Enum,
	None,
]


@lru_cache(maxsize=None)
def try_bool(provider: PROVIDER, value: bool, _f: str = "") -> str:
	if provider.startswith("sqlserver") or provider.startswith("mysql"):
		return "1" if value else "0"
	return str(value).upper()


@lru_cache(maxsize=None)
def try_str(provider: PROVIDER, value: str, _f: str = "") -> str:
	value = value.replace("'", "''")
	return f"'{value}'"


@lru_cache(maxsize=None)
def try_int(_p: PROVIDER, value: int, _f: str = "") -> int:
	return value


def try_json(provider: PROVIDER, value: Union[Dict[str, Any], List[Dict[str, Any]]], _f: str = "") -> str:
	data = json.dumps(value, ensure_ascii=False, cls=PySQLXJsonEnconder).replace("'", "''")
	return f"'{data}'"


@lru_cache(maxsize=None)
def try_uuid(_a: PROVIDER, value: UUID, _f: str = "") -> str:
	return f"'{value}'"


@lru_cache(maxsize=None)
def try_time(_p: PROVIDER, value: time, _f: str = "") -> str:
	return f"'{value}'"


@lru_cache(maxsize=None)
def try_date(_p: PROVIDER, value: date, _f: str = "") -> str:
	return f"'{value}'"


@lru_cache(maxsize=None)
def try_datetime(provider: PROVIDER, value: datetime, _f: str = "") -> str:
	if provider == "sqlserver":
		return f"'{value.isoformat(timespec='milliseconds').split('+')[0]}'"
	return f"'{value}'"


@lru_cache(maxsize=None)
def try_float(_p: PROVIDER, value: float, _f: str = "") -> float:
	return value


@lru_cache(maxsize=None)
def try_bytes(provider: PROVIDER, value: bytes, field: str = "") -> str:
	if provider == "sqlserver":
		return f"0x{value.hex()}"

	elif provider == "postgresql":
		return f"'\\x{value.hex()}'"

	elif provider == "mysql":
		return f"X'{value.hex()}'"

	# sqlite
	return f"x'{value.hex()}'"


@lru_cache(maxsize=None)
def try_decimal(_p: PROVIDER, value: Decimal, _f: str = "") -> str:
	return f"'{str(value)}'"


@lru_cache(maxsize=None)
def try_enum(provider: PROVIDER, value: Enum, field: str = "") -> str:
	new_value = value.value
	func = get_method(typ=type(new_value))
	return func(provider, new_value, field)


def try_tuple_enum(provider: PROVIDER, values: List[Enum], field: str = "") -> str:
	if not provider.startswith("postgresql"):
		raise ParameterInvalidProviderError(field=field, provider=provider, typ="array")

	types = set([type(v) for v in values])

	if len(types) > 1:
		raise ParameterInvalidValueError(
			field=field,
			provider=provider,
			typ_from="list[enum]",
			typ_to="array",
			details=(
				"the list[enum] must be of the same type, eg: [Enum1, Enum2, Enum3]. "
				"list[enum] is represented as a array in sql, sql does not support heterogeneous array."
			),
		)

	data = ",".join([str(v.value) for v in values])
	return "'{" + data + "}'"


@lru_cache(maxsize=None)
def try_tuple(provider: PROVIDER, values: Tuple[Any], field: str = "") -> str:
	types = set([type(v) for v in values])
	if len(values) > 0:
		typ_ = types.pop()
		method = get_method(typ=typ_)

		data = str([method(provider, value, field) for value in values]).replace("[", "{").replace("]", "}")
		return f"'{data}'"

	return "'{}'"


def get_method(typ: Type) -> Callable:
	METHODS = {
		bool: try_bool,
		str: try_str,  # try_str, change to try_nstr
		int: try_int,
		list: try_json,  # try_json, change to try_njson
		dict: try_json,  # try_json, change to try_njson
		tuple: try_tuple,
		UUID: try_uuid,
		time: try_time,
		date: try_date,
		datetime: try_datetime,
		float: try_float,
		bytes: try_bytes,
		Decimal: try_decimal,
	}

	return METHODS.get(typ)
